#include <queue>
#include <stack>
#include <fstream>
#include <iostream>

#include "BvhLoader.h"
#include "gzstream.h"


#ifdef _CRT_SET
	#define _CRTDBG_MAP_ALLOC
	#include <stdlib.h>
	#include <crtdbg.h>

	// redefine new operator
	#define DEBUG_NEW new(_NORMAL_BLOCK, __FILE__, __LINE__)
	#define new DEBUG_NEW
#endif


namespace CHCDemoEngine
{

using namespace std;


#define TYPE_INTERIOR -2
#define TYPE_LEAF -3



BvhNode *BvhLoader::LoadNextNode(igzstream &stream, BvhInterior *parent)
{
	int nodeType;
	stream.read(reinterpret_cast<char *>(&nodeType), sizeof(int));

	BvhNode *node;

	if (nodeType == TYPE_LEAF)
		node = new BvhLeaf(parent);
	else if (nodeType == TYPE_INTERIOR)
		node = new BvhInterior(parent);
	else
		cerr << "error: wrong node type: " << nodeType << endl;
		

	Vector3 bMin, bMax;

	stream.read(reinterpret_cast<char *>(&(node->mFirst)), sizeof(int));
	stream.read(reinterpret_cast<char *>(&(node->mLast)), sizeof(int));

	stream.read(reinterpret_cast<char *>(&bMin), sizeof(Vector3));
	stream.read(reinterpret_cast<char *>(&bMax), sizeof(Vector3));

	node->mBox = AxisAlignedBox3(bMin, bMax);
	node->mArea = node->mBox.SurfaceArea();

	//cout << "box: " << node->mBox << " area: " << node->mArea << endl;
	
	return node;
}


Bvh *BvhLoader::Load(const string &filename,  
					 const SceneEntityContainer &staticEntities,
					 const SceneEntityContainer &dynamicEntities)
{
	queue<BvhNode *> tQueue;
	igzstream stream(filename.c_str());

	if (!stream.is_open()) return NULL;

	cout << "loading bvh" << endl;

	Bvh *bvh = new Bvh(staticEntities, dynamicEntities);

	BvhNode *root = LoadNextNode(stream, NULL);

	bvh->mStaticRoot = root;
	bvh->mNumNodes = 3;

	tQueue.push(root);
	
	while(!tQueue.empty())
	{
		BvhNode *node = tQueue.front();
		tQueue.pop();

		if (!node->IsLeaf())
		{
			bvh->mNumNodes += 2;

			BvhInterior *interior = static_cast<BvhInterior *>(node);

			BvhNode *front = LoadNextNode(stream, interior);
			BvhNode *back = LoadNextNode(stream, interior);

			interior->mFront = front;
			interior->mBack = back;

			front->mDepth = interior->mDepth + 1;
			back->mDepth = interior->mDepth + 1;

			tQueue.push(front);			
			tQueue.push(back);
		}
	}


	////////////
	//-- create dynamic part of the hierarchy

	bvh->CreateDynamicBranch();

	
	///////////
	//-- post process nodes

	/// this function must be called once after creation
	bvh->PostProcess();


	cout << "... finished loading " << bvh->mNumNodes << " nodes" << endl;
	cout << "scene box: " << bvh->mBox << endl;

	return bvh;
}


}
